//=============================================================================
// File: DLCStore.gs
// Desc: Declares the classes for the in game content store
//=============================================================================
include "SecurityToken.gs"


//=============================================================================
// Name: DLCInfo
// Desc: Represents a single content package
//=============================================================================
final game class DLCInfo isclass GSObject
{

  //=============================================================================
  // Name: GeProductID
  // Desc: Returns a unique identifier for this product
  //=============================================================================
  public native string GetProductID();


  //=============================================================================
  // Name: GetLocalisedName
  // Desc: Returns a localised display name for this product
  //=============================================================================
  public native string GetLocalisedName();


  //=============================================================================
  // Name: GetLocalisedDescription
  // Desc: Returns a localised description for this product
  //=============================================================================
  public native string GetLocalisedDescription();


  //=============================================================================
  // Name: GetLocalisedPrice
  // Desc: Returns the cost of this product in a user-readable format
  //=============================================================================
  public native string GetLocalisedPrice();


  //=============================================================================
  // Name: GetIconPath
  // Desc: Returns the kuid for a small icon image, suitable for setting as a
  //       html img kuid param.
  //=============================================================================
  public native string GetIconPath();


  //=============================================================================
  // Name: GetThumbnailPath
  // Desc: Returns the kuid for a large thumbnail image, suitable for setting as
  //       a html img kuid param.
  //=============================================================================
  public native string GetThumbnailPath();


  //=============================================================================
  // Name: IsInstalled
  // Desc: Returns true if the DLC package has been purchased and is locally
  //       installed.
  //=============================================================================
  public native bool IsLocal();


  //=============================================================================
  // Name: IsPurchased
  // Desc: Returns true if the DLC package has been purchased. The content may or
  //       may not be locally available, see DLCInfo.IsLocal()
  //=============================================================================
  public native bool IsPurchased();

};



//=============================================================================
// Name: DLCStore
// Desc: The in game DLC store. Allows purchasing of content in game.
//
//       Posts the following messages:
//  | Message.major         | Message.minor                                   |
//  | "DLCStore"            | "product-list-changed"                          |
//  | "DLCStore"            | "product-restore-succeeded"                     |
//  | "DLCStore"            | "product-restore-failed"                        |
//  | "DLCStore"            | "product-register-failed"                       |
//  | "DLCStoreTransaction" | "<product id>-<transaction status>"             |
//    Where product id is identical to DLCInfo.GetProductID() and status is one
//    of the following: "unknown", "purchasing", "purchased", "failed"
//=============================================================================
final static class DLCStore isclass GameObject
{

  //=============================================================================
  // Name: GetProductList
  // Desc: Get the full list of purchasable content
  //=============================================================================
  public native DLCInfo[] GetProductList(SecurityToken token, GSObject caller);


  //=============================================================================
  // Name: GetProductInfo
  // Desc: Get the DLCInfo for a specific product
  //=============================================================================
  public native DLCInfo GetProductInfo(SecurityToken token, GSObject caller, string productId);


  //=============================================================================
  // Name: IsPurchasingEnabled
  // Desc: Returns whether purchasing is enabled
  //=============================================================================
  public native bool IsPurchasingEnabled(SecurityToken token, GSObject caller);


  //=============================================================================
  // Name: SubmitPurchase
  // Desc: Submits a purchase to the store
  //=============================================================================
  public native void SubmitPurchase(SecurityToken token, GSObject caller, string productId);


  //=============================================================================
  // Name: RedownloadAllPurchases
  // Desc: Redownloads every product purchased by this player
  //=============================================================================
  public native void RedownloadAllPurchases(SecurityToken token, GSObject caller);

};

